import sys

import pytest

from easyprocess import EasyProcess

python = sys.executable


def test_timeout():
    p = EasyProcess("sleep 1").start()
    p.wait(0.2)
    assert p.is_alive()
    p.wait(0.2)
    assert p.is_alive()
    p.wait(2)
    assert not p.is_alive()

    assert EasyProcess("sleep 0.3").call().return_code == 0
    assert EasyProcess("sleep 0.3").call(timeout=0.1).return_code != 0
    assert EasyProcess("sleep 0.3").call(timeout=1).return_code == 0

    assert EasyProcess("sleep 0.3").call().timeout_happened is False
    assert EasyProcess("sleep 0.3").call(timeout=0.1).timeout_happened
    assert EasyProcess("sleep 0.3").call(timeout=1).timeout_happened is False


@pytest.mark.timeout(3)
def test_time_cli1():
    p = EasyProcess(
        [
            python,
            "-c",
            "import logging;logging.basicConfig(level=logging.DEBUG);from easyprocess import EasyProcess;EasyProcess('sleep 5').start()",
        ]
    )
    p.call()
    assert p.return_code == 0


@pytest.mark.timeout(3)
def test_time_cli2():
    p = EasyProcess(
        [
            python,
            "-c",
            "import logging;logging.basicConfig(level=logging.DEBUG);from easyprocess import EasyProcess;EasyProcess('sleep 5').call(timeout=0.5)",
        ]
    )
    p.call()
    assert p.return_code == 0


@pytest.mark.timeout(3)
def test_time2():
    p = EasyProcess("sleep 5").call(timeout=1)
    assert p.is_alive() is False
    assert p.timeout_happened
    assert p.return_code != 0
    assert p.stdout == ""


@pytest.mark.timeout(3)
def test_timeout_out():
    p = EasyProcess(
        [python, "-c", "import time;print( 'start');time.sleep(5);print( 'end')"]
    ).call(timeout=1)
    assert p.is_alive() is False
    assert p.timeout_happened
    assert p.return_code != 0
    assert p.stdout == ""


@pytest.mark.timeout(0.3)
def test_time3():
    EasyProcess("sleep 5").start()



import logging
import shlex
import sys
import unicodedata

log = logging.getLogger(__name__)

PY3 = sys.version_info[0] >= 3

if PY3:
    string_types = (str,)
else:
    string_types = (basestring,)


class EasyProcessUnicodeError(Exception):
    pass


def split_command(cmd, posix=None):
    """
     - cmd is string list -> nothing to do
     - cmd is string -> split it using shlex

    :param cmd: string ('ls -l') or list of strings (['ls','-l'])
    :rtype: string list
    """
    if not isinstance(cmd, string_types):
        # cmd is string list
        pass
    else:
        if not PY3:
            # cmd is string
            # The shlex module currently does not support Unicode input (in
            # 2.x)!
            if isinstance(cmd, unicode):
                try:
                    cmd = unicodedata.normalize("NFKD", cmd).encode("ascii", "strict")
                except UnicodeEncodeError:
                    raise EasyProcessUnicodeError(
                        'unicode command "%s" can not be processed.' % cmd
                        + "Use string list instead of string"
                    )
                log.debug("unicode is normalized")
        if posix is None:
            posix = True
        cmd = shlex.split(cmd, posix=posix)
    return cmd


def uniencode(s):
    if PY3:
        pass
    #        s=s.encode()
    else:
        if isinstance(s, unicode):
            s = s.encode("utf-8")
    return s


def unidecode(s):
    if PY3:
        s = s.decode("utf-8", "ignore")
    else:
        if isinstance(s, str):
            s = s.decode("utf-8", "ignore")
    return s



"""Easy to use python subprocess interface."""

import logging
import os.path
import signal
import subprocess
import tempfile
import threading
import time

from easyprocess.about import __version__
from easyprocess.unicodeutil import split_command, unidecode, uniencode

log = logging.getLogger(__name__)

log.debug("version=%s", __version__)

POLL_TIME = 0.1
USE_POLL = 0


class EasyProcessError(Exception):
    def __init__(self, easy_process, msg=""):
        self.easy_process = easy_process
        self.msg = msg

    def __str__(self):
        return self.msg + " " + repr(self.easy_process)


template = """cmd=%s
OSError=%s
Program install error! """


class EasyProcess(object):

    """
    .. module:: easyprocess

    simple interface for :mod:`subprocess`

    shell is not supported (shell=False)

    .. warning::

      unicode is supported only for string list command (Python2.x)
      (check :mod:`shlex` for more information)

    :param cmd: string ('ls -l') or list of strings (['ls','-l'])
    :param cwd: working directory
    :param use_temp_files: use temp files instead of pipes for
                           stdout and stderr,
                           pipes can cause deadlock in some cases
                           (see unit tests)

    :param env: If *env* is not ``None``, it must be a mapping that defines the environment
                   variables for the new process; these are used instead of inheriting the current
                   process' environment, which is the default behavior.
                   (check :mod:`subprocess`  for more information)
    """

    def __init__(
        self, cmd, cwd=None, use_temp_files=True, env=None,
    ):
        self.use_temp_files = use_temp_files
        self._outputs_processed = False

        self.env = env
        self.popen = None
        self.stdout = None
        self.stderr = None
        self._stdout_file = None
        self._stderr_file = None
        self.is_started = False
        self.oserror = None
        self.cmd_param = cmd
        self._thread = None
        self.timeout_happened = False
        self.cwd = cwd
        cmd = split_command(cmd)
        self.cmd = cmd
        # self.cmd_as_string = " ".join(self.cmd)
        self.enable_stdout_log = True
        self.enable_stderr_log = True

        # log.debug('param: "%s" ', self.cmd_param)
        log.debug("command: %s", self.cmd)
        # log.debug('joined command: %s', self.cmd_as_string)

        if not len(cmd):
            raise EasyProcessError(self, "empty command!")

    def __repr__(self):
        msg = (
            '<%s cmd_param=%s cmd=%s oserror=%s return_code=%s stdout="%s" stderr="%s" timeout_happened=%s>'
            % (
                self.__class__.__name__,
                self.cmd_param,
                self.cmd,
                self.oserror,
                self.return_code,
                self.stdout,
                self.stderr,
                self.timeout_happened,
            )
        )
        return msg

    @property
    def pid(self):
        """
        PID (:attr:`subprocess.Popen.pid`)

        :rtype: int
        """
        if self.popen:
            return self.popen.pid

    @property
    def return_code(self):
        """
        returncode (:attr:`subprocess.Popen.returncode`)

        :rtype: int
        """
        if self.popen:
            return self.popen.returncode

    def call(self, timeout=None):
        """Run command with arguments. Wait for command to complete.

        same as:
         1. :meth:`start`
         2. :meth:`wait`
         3. :meth:`stop`

        :rtype: self

        """
        self.start().wait(timeout=timeout)
        if self.is_alive():
            self.stop()
        return self

    def start(self):
        """start command in background and does not wait for it.

        :rtype: self

        """
        if self.is_started:
            raise EasyProcessError(self, "process was started twice!")

        if self.use_temp_files:
            self._stdout_file = tempfile.TemporaryFile(prefix="stdout_")
            self._stderr_file = tempfile.TemporaryFile(prefix="stderr_")
            stdout = self._stdout_file
            stderr = self._stderr_file

        else:
            stdout = subprocess.PIPE
            stderr = subprocess.PIPE

        cmd = list(map(uniencode, self.cmd))

        try:
            self.popen = subprocess.Popen(
                cmd, stdout=stdout, stderr=stderr, cwd=self.cwd, env=self.env,
            )
        except OSError as oserror:
            log.debug("OSError exception: %s", oserror)
            self.oserror = oserror
            raise EasyProcessError(self, "start error")
        self.is_started = True
        log.debug("process was started (pid=%s)", self.pid)
        return self

    def is_alive(self):
        """
        poll process using :meth:`subprocess.Popen.poll`
        It updates stdout/stderr/return_code if process has stopped earlier.
        
        :rtype: bool
        """
        if self.popen:
            alive = self.popen.poll() is None

            if not alive:
                # collect stdout/stderr/return_code if proc stopped
                self._wait4process()

            return alive
        else:
            return False

    def wait(self, timeout=None):
        """Wait for command to complete.

        Timeout:
         - discussion: http://stackoverflow.com/questions/1191374/subprocess-with-timeout
         - implementation: threading

        :rtype: self

        """

        if timeout is not None:
            if not self._thread:
                self._thread = threading.Thread(target=self._wait4process)
                self._thread.daemon = 1
                self._thread.start()

        if self._thread:
            self._thread.join(timeout=timeout)
            self.timeout_happened = self.timeout_happened or self._thread.is_alive()
        else:
            # no timeout and no existing thread
            self._wait4process()

        return self

    def _wait4process(self):
        if self._outputs_processed:
            return

        def remove_ending_lf(s):
            if s.endswith("\n"):
                s = s[:-1]
            if s.endswith("\r"):
                s = s[:-1]
            return s

        if self.popen:
            if self.use_temp_files:
                if USE_POLL:
                    while True:
                        if self.popen.poll() is not None:
                            break
                        # if self._stop_thread:
                        #     return
                        time.sleep(POLL_TIME)

                else:
                    # wait() blocks process, timeout not possible
                    self.popen.wait()

                self._stdout_file.seek(0)
                self._stderr_file.seek(0)
                self.stdout = self._stdout_file.read()
                self.stderr = self._stderr_file.read()

                self._stdout_file.close()
                self._stderr_file.close()
            else:
                # This will deadlock when using stdout=PIPE and/or stderr=PIPE
                # and the child process generates enough output to a pipe such
                # that it blocks waiting for the OS pipe buffer to accept more data.
                # Use communicate() to avoid that.
                # self.popen.wait()
                # self.stdout = self.popen.stdout.read()
                # self.stderr = self.popen.stderr.read()

                # communicate() blocks process, timeout not possible
                (self.stdout, self.stderr) = self.popen.communicate()
            log.debug("process has ended, return code=%s", self.return_code)
            self.stdout = remove_ending_lf(unidecode(self.stdout))
            self.stderr = remove_ending_lf(unidecode(self.stderr))
            self._outputs_processed = True

            #            def limit_str(s):
            #                if len(s) > self.max_bytes_to_log:
            #                    warn = '[middle of output was removed, max_bytes_to_log=%s]'%(self.max_bytes_to_log)
            #                    s = s[:self.max_bytes_to_log / 2] + warn + s[-self.max_bytes_to_log / 2:]
            #                return s
            if self.enable_stdout_log:
                log.debug("stdout=%s", self.stdout)
            if self.enable_stderr_log:
                log.debug("stderr=%s", self.stderr)

    def stop(self):
        """Kill process and wait for command to complete.

        same as:
         1. :meth:`sendstop`
         2. :meth:`wait`

        :rtype: self

        """
        return self.sendstop().wait()

    def sendstop(self):
        """
        Kill process (:meth:`subprocess.Popen.terminate`).
        Do not wait for command to complete.

        :rtype: self
        """
        if not self.is_started:
            raise EasyProcessError(self, "process was not started!")

        log.debug('stopping process (pid=%s cmd="%s")', self.pid, self.cmd)
        if self.popen:
            if self.is_alive():
                log.debug("process is active -> sending SIGTERM")

                try:
                    try:
                        self.popen.terminate()
                    except AttributeError:
                        os.kill(self.popen.pid, signal.SIGKILL)
                except OSError as oserror:
                    log.debug("exception in terminate:%s", oserror)

            else:
                log.debug("process was already stopped")
        else:
            log.debug("process was not started")

        return self

    def sleep(self, sec):
        """
        sleeping (same as :func:`time.sleep`)

        :rtype: self
        """
        time.sleep(sec)

        return self

    def wrap(self, func, delay=0):
        """
        returns a function which:
         1. start process
         2. call func, save result
         3. stop process
         4. returns result

        similar to :keyword:`with` statement

        :rtype:
        """

        def wrapped():
            self.start()
            if delay:
                self.sleep(delay)
            x = None
            try:
                x = func()
            except OSError as oserror:
                log.debug("OSError exception:%s", oserror)
                self.oserror = oserror
                raise EasyProcessError(self, "wrap error!")
            finally:
                self.stop()
            return x

        return wrapped

    def __enter__(self):
        """used by the :keyword:`with` statement"""
        self.start()
        return self

    def __exit__(self, *exc_info):
        """used by the :keyword:`with` statement"""
        self.stop()



import sys

import pytest
import six

from easyprocess import EasyProcess
from easyprocess.unicodeutil import EasyProcessUnicodeError, split_command

u = six.u
OMEGA = u("\u03A9")

python = sys.executable


def test_str():
    assert EasyProcess(u("ls -la")).call().return_code == 0


def test_ls():
    assert EasyProcess([u("ls"), u("-la")]).call().return_code == 0


def test_parse():
    assert EasyProcess(u("ls -la")).cmd == ["ls", "-la"]
    assert EasyProcess(u('ls "abc"')).cmd == ["ls", "abc"]
    assert EasyProcess(u('ls "ab c"')).cmd == ["ls", "ab c"]


def test_split():
    # list -> list
    assert split_command([str("x"), str("y")]) == ["x", "y"]
    assert split_command([str("x"), u("y")]) == ["x", "y"]
    assert split_command([str("x"), OMEGA]) == ["x", OMEGA]

    # str -> list
    assert split_command(str("x y")) == ["x", "y"]
    assert split_command(u("x y")) == ["x", "y"]
    if six.PY3:
        assert split_command(u("x ") + OMEGA) == ["x", OMEGA]
    else:
        with pytest.raises(EasyProcessUnicodeError):
            split_command(u("x ") + OMEGA)

    # split windows paths #12
    assert (
        split_command("c:\\temp\\a.exe someArg", posix=False)
        == ["c:\\temp\\a.exe", "someArg"],
    )


def test_echo():
    assert EasyProcess(u("echo hi")).call().stdout == "hi"

    if six.PY3:
        assert EasyProcess(u("echo ") + OMEGA).call().stdout == OMEGA
    else:
        # unicode is not supported
        with pytest.raises(EasyProcessUnicodeError):
            EasyProcess(u("echo ") + OMEGA).call().stdout

    assert EasyProcess(["echo", OMEGA]).call().stdout == OMEGA


def test_invalid_stdout():
    """invalid utf-8 byte in stdout."""
    # https://en.wikipedia.org/wiki/UTF-8#Codepage_layout

    # 0x92  continuation byte
    if six.PY3:
        cmd = [python, "-c", "import sys;sys.stdout.buffer.write(b'\\x92')"]
    else:
        cmd = [python, "-c", "import sys;sys.stdout.write(b'\\x92')"]
    p = EasyProcess(cmd).call()
    assert p.return_code == 0
    assert p.stdout == ""

    # 0xFF must never appear in a valid UTF-8 sequence
    if six.PY3:
        cmd = [python, "-c", "import sys;sys.stdout.buffer.write(b'\\xFF')"]
    else:
        cmd = [python, "-c", "import sys;sys.stdout.write(b'\\xFF')"]
    p = EasyProcess(cmd).call()
    assert p.return_code == 0
    assert p.stdout == ""



import sys

from easyprocess import EasyProcess

python = sys.executable

print("-- Run program, wait for it to complete, get stdout:")
s = EasyProcess([python, "-c", "print(3)"]).call().stdout
print(s)

print("-- Run program, wait for it to complete, get stderr:")
s = EasyProcess([python, "-c", "import sys;sys.stderr.write('4\\n')"]).call().stderr
print(s)

print("-- Run program, wait for it to complete, get return code:")
s = EasyProcess([python, "--version"]).call().return_code
print(s)

print("-- Run program, wait 1 second, stop it, get stdout:")
s = EasyProcess(["ping", "localhost"]).start().sleep(1).stop().stdout
print(s)



import sys

import pytest

from easyprocess import EasyProcess
from pyvirtualdisplay.display import Display

python = sys.executable

VISIBLE = 0

# deadlock in 0.0.0
# popen.communicate() hangs
# no deadlock with temp_files


@pytest.mark.timeout(100)
def test_deadlock():
    # skip these tests for Windows/Mac
    if not sys.platform.startswith("linux"):
        return
    d = Display(visible=VISIBLE, size=(600, 400))
    d.start()

    p = EasyProcess([python, "-c", 'import Image;Image.new("RGB",(99, 99)).show()'])
    p.start()
    p.sleep(1)
    # hangs with pipes
    p.stop()

    d.stop()



import os

from setuptools import setup

NAME = "easyprocess"
PYPI_NAME = "EasyProcess"
URL = "https://github.com/ponty/easyprocess"
DESCRIPTION = "Easy to use Python subprocess interface."
LONG_DESCRIPTION = """Easy to use Python subprocess interface.

home: https://github.com/ponty/easyprocess"""
PACKAGES = [
    NAME,
    NAME + ".examples",
]

# get __version__
__version__ = None
exec(open(os.path.join(NAME, "about.py")).read())
VERSION = __version__

# extra = {}
# if sys.version_info >= (3,):
#     extra['use_2to3'] = True

classifiers = [
    # Get more strings from
    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    "License :: OSI Approved :: BSD License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
]


setup(
    name=PYPI_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    classifiers=classifiers,
    keywords="subprocess interface",
    author="ponty",
    # author_email='',
    url=URL,
    license="BSD",
    packages=PACKAGES,
    # **extra
)
